"""TeamCapacity."""
import pandas as pd
from .team import Team
from .holiday import HolidaySchedulePort
from .person import Person
from .epic import EpicType
from .timeperiod import TimePeriod

class TeamCapacity():
    # pylint: disable=line-too-long, too-many-instance-attributes
    """Class representing the capacity for a team."""
    def __init__(self,  team: Team, time_period:TimePeriod, holiday_schedule:HolidaySchedulePort) -> None:
        if time_period.is_valid() is False:
            raise ValueError('Invalid time period')
        self.team = team
        self.start_date = time_period.start_date
        self.end_date = time_period.end_date
        self.date_range = pd.date_range(time_period.start_date,time_period.end_date, freq="D")
        self.holiday_schedule = holiday_schedule
        self.total_capacity_data = {}
        self.leading_static_column_headings = [
            'Team',
            'Person',
            'Location',
            'Start Date',
            'End Date',
            'Front End',
            'Back End',
            'QE',
            'Documentation',
            'DevOps',
            'Reserve Capacity'
        ]
        self.daily_column_headings = [dt.strftime('%Y-%m-%d') for dt in self.date_range]
        self.trailing_static_column_headings = [
            'Total'
        ]
        self.data = {}
        all_column_headings = self.leading_static_column_headings + self.daily_column_headings + self.trailing_static_column_headings 
        for column_heading in all_column_headings:
            self.data[column_heading] = []
        self.detailed_capacity_data = {}
        self.df = None

    def is_person_unavailable(self,person):
        """Returns true if the person is unavailable for the time period."""
        if person.start_date > self.end_date:
            return True
        if person.end_date < self.start_date:
            return True
        return False

    def populate_leading_static_columns_for_person(self,person:Person):
        """Populates the leading static columns for a person."""
        self.data['Team'].append(self.team.name)
        self.data['Person'].append(person.name)
        self.data['Location'].append(person.location)
        self.data['Start Date'].append(person.start_date.isoformat())
        self.data['End Date'].append(person.end_date.isoformat())
        self.data['Front End'].append('T' if person.front_end else 'F')
        self.data['Back End'].append('T' if person.back_end else 'F')
        self.data['QE'].append('T' if person.qe else 'F')
        self.data['DevOps'].append('T' if person.devops else 'F')
        self.data['Documentation'].append('T' if person.documentation else 'F')
        self.data['Reserve Capacity'].append(person.reserve_capacity)

    def populate_daily_columns_with_zeros_for_person(self,person):
        """Populates the daily columns with zeros for a person."""
        for daily_column in self.daily_column_headings:
            self.data[daily_column].append(0)
            self.populate_detailed_capacity(daily_column,person,0)
        self.data['Total'].append(0)

    def populate_detailed_capacity(self,for_day,person,capacity_for_person_for_this_day):
        """Populates the detailed capacity for a person for a given day."""
        detailed_capacity_data_for_day = self.detailed_capacity_data.setdefault(for_day, {})
        detailed_capacity_data_for_day_for_person = detailed_capacity_data_for_day.setdefault(person.name, {}) 
        detailed_capacity_data_for_day_for_person[EpicType.FRONTEND.name] = [
            capacity_for_person_for_this_day if person.front_end else 0]
        detailed_capacity_data_for_day_for_person[EpicType.BACKEND.name] = [
            capacity_for_person_for_this_day if person.back_end else 0]
        detailed_capacity_data_for_day_for_person[EpicType.QE.name] = [
            capacity_for_person_for_this_day if person.qe else 0]
        detailed_capacity_data_for_day_for_person[EpicType.DEVOPS.name] = [
            capacity_for_person_for_this_day if person.devops else 0]
        detailed_capacity_data_for_day_for_person[EpicType.DOCUMENTATION.name] = [
            capacity_for_person_for_this_day if person.documentation else 0]

    def its_the_weekend(self,dayofweek):
        """Returns true if the given day of the week is a weekend."""
        return dayofweek > 4

    def its_a_holiday(self, some_date, location):
        """Returns true if the given date is a holiday."""
        return self.holiday_schedule.falls_on_holiday(some_date,location)

    def populate_daily_columns_for_person(self,person):
        """Populates the daily columns for a person."""
        total = 0
        for dt in self.date_range:
            daily_column = dt.strftime('%Y-%m-%d')
            if (person.is_not_available(dt.date()) or
                 self.its_the_weekend(dt.dayofweek) or
                 self.its_a_holiday(dt.date(),person.location)):
                self.data[daily_column].append(0)
                self.populate_detailed_capacity(daily_column,person,0)
            else:
                capacity_for_person_for_this_day = (1.0-person.reserve_capacity) * 1.0
                self.data[daily_column].append(capacity_for_person_for_this_day)
                self.populate_detailed_capacity(daily_column,person,capacity_for_person_for_this_day)
                total += capacity_for_person_for_this_day 
        self.data['Total'].append(total)

    def populate_total_row(self):
        """Populates the total row."""
        for leading_static_column in self.leading_static_column_headings:
            if leading_static_column == 'Team':
                self.data[leading_static_column].append(self.team.name)
            elif leading_static_column == 'Person':
                self.data[leading_static_column].append('Total')
            else:
                self.data[leading_static_column].append('-')
        for dt in self.date_range:
            daily_column = dt.strftime('%Y-%m-%d')
            daily_total = sum(self.data[daily_column])
            self.data[daily_column].append(daily_total)
            self.total_capacity_data[daily_column] = [daily_total] 
        team_total = sum(self.data['Total'])
        self.data['Total'].append(team_total)

    def calculate(self):
        """Calculates the capacity for the team."""
        for person in self.team.person_list:
            self.populate_leading_static_columns_for_person(person)
            if self.is_person_unavailable(person):
                self.populate_daily_columns_with_zeros_for_person(person)
            else:
                self.populate_daily_columns_for_person(person)
        self.populate_total_row()
        self.df = pd.DataFrame(self.data)

    def get_df(self) -> pd.DataFrame:
        """Returns the capacity data as a pandas dataframe."""
        return self.df



"""Unit tests for scheduler.py"""
import unittest
from datetime import date
from ..src.scheduler import TeamScheduler, EpicScheduleStatus
from ..src.epic import Epic, EpicType
from ..src.teamcapacity import TeamCapacity
from ..src.team import Team
from ..src.timeperiod import TimePeriod
from ..src.holiday import HolidaySchedulePort

class HolidayScheduleForTesting(HolidaySchedulePort):
    """HolidaySchedulePort implementation for testing."""
    # overriding abstract method
    def falls_on_holiday(self,some_date: date,location: str) -> bool:
        return False

class TestTeamScheduler(unittest.TestCase):
    """Tests for TeamScheduler"""

    def test_schedule_epic_team_unavailable(self):
        """Tests scheduling an epic when the team is unavailable."""
        team1_document = """
        team:
          name: Team1
          persons:
          - name: Freddy UIDev
            start_date: '2023-01-01'
            end_date: '2023-01-31'
            front_end: True
            back_end: True
            qe: False
            devops: False
            documentation: False
            reserve_capacity: 0.0
            location: US
            out_of_office_dates: []
        """
        time_period = TimePeriod(
            name='test_period',
            start_date=date(2023,10,25),
            end_date=date(2023,11,7)
        )
        e1 = Epic(
            key='csesc-1050',
            estimated_size=2,
            epic_type=EpicType.FRONTEND
        )
        team1 = Team('Team1', 'team1.yaml').load_from_yaml_as_string(team1_document)
        team1_capacity = TeamCapacity(
            team1,
            time_period,
            HolidayScheduleForTesting())
        team1_capacity.calculate()
        scheduler = TeamScheduler(team1_capacity,[e1])
        schedule_results = scheduler.build_schedule()
        self.assertEqual(len(schedule_results), 1)
        self.assertEqual(schedule_results[0].start_date, 'WILL NOT START')
        self.assertEqual(schedule_results[0].end_date, 'WILL NOT COMPLETE IN TIME')

    def test_schedule_epic_that_fits(self):
        """Tests scheduling an epic that fits."""
        team1_document = """
        team:
          name: Team1
          persons:
          - name: Freddy UIDev
            start_date: '2023-01-01'
            end_date: '2030-12-31'
            front_end: True
            back_end: True
            qe: False
            devops: False
            documentation: False
            reserve_capacity: 0.0
            location: US
            out_of_office_dates: []
        """
        time_period = TimePeriod(
            name='test_period',
            start_date=date(2023,10,25),
            end_date=date(2023,11,7)
        )
        e1 = Epic(
            key='csesc-1050',
            estimated_size=2,
            epic_type=EpicType.FRONTEND
        )
        team1 = Team('Team1', 'team1.yaml').load_from_yaml_as_string(team1_document)
        team1_capacity = TeamCapacity(
            team1, 
            time_period,
            HolidayScheduleForTesting())
        team1_capacity.calculate()
        scheduler = TeamScheduler(team1_capacity,[e1])
        schedule_results = scheduler.build_schedule()
        self.assertEqual(len(schedule_results), 1)
        self.assertEqual(schedule_results[0].start_date, '2023-10-25')
        self.assertEqual(schedule_results[0].end_date, '2023-10-26')
        self.assertEqual(schedule_results[0].epic_schedule_status, EpicScheduleStatus.OK)

    def test_schedule_epic_but_team_doesnt_have_skill_match(self):
        """Tests scheduling an epic when the team doesn't have a skill match."""
        team1_document = """
        team:
          name: Team1
          persons:
          - name: Freddy UIDev
            start_date: '2023-01-01'
            end_date: '2030-12-31'
            front_end: True
            back_end: True
            qe: False
            devops: False
            documentation: False
            reserve_capacity: 0.0
            location: US
            out_of_office_dates: []
        """
        time_period = TimePeriod(
            name='test_period',
            start_date=date(2023,10,25),
            end_date=date(2023,11,7)
        )
        e1 = Epic(
            key='csesc-1050',
            estimated_size=2,
            epic_type=EpicType.DOCUMENTATION
        )
        team1 = Team('Team1', 'team1.yaml').load_from_yaml_as_string(team1_document)
        team1_capacity = TeamCapacity(
            team1,
            time_period,
            HolidayScheduleForTesting())
        team1_capacity.calculate()
        scheduler = TeamScheduler(team1_capacity,[e1])
        schedule_results = scheduler.build_schedule()
        self.assertEqual(len(schedule_results), 1)
        self.assertEqual(schedule_results[0].start_date, 'WILL NOT START')
        self.assertEqual(schedule_results[0].end_date, 'WILL NOT COMPLETE IN TIME')
        self.assertEqual(schedule_results[0].epic_schedule_status, EpicScheduleStatus.NO_CAPACITY_TO_START)

    def test_schedule_epic_make_sure_person_is_not_over_allocated(self):
        """Tests scheduling an epic when the person is over allocated."""
        team1_document = """
        team:
          name: Team1
          persons:
          - name: Freddy Can Do It All
            start_date: '2023-01-01'
            end_date: '2030-12-31'
            front_end: True
            back_end: True
            qe: False
            devops: False
            documentation: False
            reserve_capacity: 0.0
            location: US
            out_of_office_dates: []
        """
        time_period = TimePeriod(
            name='test_period',
            start_date=date(2023,10,25),
            end_date=date(2023,11,7)
        )
        e1 = Epic(
            key='csesc-1050',
            estimated_size=2,
            epic_type=EpicType.FRONTEND
        )
        e2 = Epic(
            key='csesc-1051',
            estimated_size=2,
            epic_type=EpicType.BACKEND
        )
        team1 = Team('Team1', 'team1.yaml').load_from_yaml_as_string(team1_document)
        team1_capacity = TeamCapacity(
            team1,
            time_period,
            HolidayScheduleForTesting())
        team1_capacity.calculate()
        scheduler = TeamScheduler(team1_capacity,[e1,e2])
        schedule_results = scheduler.build_schedule()
        self.assertEqual(len(schedule_results), 2)
        self.assertEqual(schedule_results[0].epic_key, 'csesc-1050')
        self.assertEqual(schedule_results[0].start_date, '2023-10-25')
        self.assertEqual(schedule_results[0].end_date, '2023-10-26')
        self.assertEqual(schedule_results[0].epic_schedule_status, EpicScheduleStatus.OK)
        self.assertEqual(schedule_results[1].epic_key, 'csesc-1051')
        self.assertEqual(schedule_results[1].start_date, '2023-10-27')
        self.assertEqual(schedule_results[1].end_date, '2023-10-30')
        self.assertEqual(schedule_results[1].epic_schedule_status, EpicScheduleStatus.OK)

    def test_schedule_epic_that_fits_with_team_of_two_but_only_one_has_skill_match(self):
        """Tests scheduling an epic that fits with a team of two but only one has a skill match."""
        team1_document = """
        team:
          name: Team1
          persons:
          - name: Freddy UIDev
            start_date: '2023-01-01'
            end_date: '2030-12-31'
            front_end: True
            back_end: True
            qe: False
            devops: False
            documentation: False
            reserve_capacity: 0.0
            location: US
            out_of_office_dates: []
          - name: Bobby BackendDev
            start_date: '2023-01-01'
            end_date: '2030-12-31'
            front_end: False
            back_end: True
            qe: False
            devops: False
            documentation: False
            reserve_capacity: 0.0
            location: US
            out_of_office_dates: []
        """
        time_period = TimePeriod(
            name='test_period',
            start_date=date(2023,10,25),
            end_date=date(2023,11,7)
        )
        e1 = Epic(
            key='csesc-1050',
            estimated_size=2,
            epic_type=EpicType.FRONTEND
        )
        team1 = Team('Team1', 'team1.yaml').load_from_yaml_as_string(team1_document)
        team1_capacity = TeamCapacity(
            team1,
            time_period,
            HolidayScheduleForTesting())
        team1_capacity.calculate()
        scheduler = TeamScheduler(team1_capacity,[e1])
        schedule_results = scheduler.build_schedule()
        self.assertEqual(len(schedule_results), 1)
        self.assertEqual(schedule_results[0].start_date, '2023-10-25')
        self.assertEqual(schedule_results[0].end_date, '2023-10-26')
        self.assertEqual(schedule_results[0].epic_schedule_status, EpicScheduleStatus.OK)

    def test_schedule_epic_that_fits_with_team_of_two(self):
        """Tests scheduling an epic that fits with a team of two."""
        team1_document = """
        team:
          name: Team1
          persons:
          - name: Freddy UIDev
            start_date: '2023-01-01'
            end_date: '2030-12-31'
            front_end: True
            back_end: True
            qe: False
            devops: False
            documentation: False
            reserve_capacity: 0.0
            location: US
            out_of_office_dates: []
          - name: Bobby UIDev
            start_date: '2023-01-01'
            end_date: '2030-12-31'
            front_end: True
            back_end: True
            qe: False
            devops: False
            documentation: False
            reserve_capacity: 0.0
            location: US
            out_of_office_dates: []
        """
        time_period = TimePeriod(
            name='test_period',
            start_date=date(2023,10,25),
            end_date=date(2023,11,7)
        )
        e1 = Epic(
            key='csesc-1050',
            estimated_size=2,
            epic_type=EpicType.FRONTEND
        )
        team1 = Team('Team1', 'team1.yaml').load_from_yaml_as_string(team1_document)
        team1_capacity = TeamCapacity(
            team1,
            time_period,
            HolidayScheduleForTesting())
        team1_capacity.calculate()
        scheduler = TeamScheduler(team1_capacity,[e1])
        schedule_results = scheduler.build_schedule()
        self.assertEqual(len(schedule_results), 1)
        self.assertEqual(schedule_results[0].start_date, '2023-10-25')
        self.assertEqual(schedule_results[0].end_date, '2023-10-25')
        self.assertEqual(schedule_results[0].epic_schedule_status, EpicScheduleStatus.OK)

    def test_schedule_two_epics_that_fit(self):
        """Tests scheduling two epics that fit."""
        team1_document = """
        team:
          name: Team1
          persons:
          - name: Freddy UIDev
            start_date: '2023-01-01'
            end_date: '2030-12-31'
            front_end: True
            back_end: False
            qe: False
            devops: False
            documentation: False
            reserve_capacity: 0.0
            location: US
            out_of_office_dates: []
        """
        time_period = TimePeriod(
            name='test_period',
            start_date=date(2023,10,25),
            end_date=date(2023,11,7)
        )
        e1 = Epic(
            key='csesc-1050',
            estimated_size=2,
            epic_type=EpicType.FRONTEND
        )
        e2 = Epic(
            key='csesc-1051',
            estimated_size=2,
            epic_type=EpicType.FRONTEND
        )
        team1 = Team('Team1', 'team1.yaml').load_from_yaml_as_string(team1_document)
        team1_capacity = TeamCapacity(
            team1,
            time_period,
            HolidayScheduleForTesting())
        team1_capacity.calculate()
        scheduler = TeamScheduler(team1_capacity,[e1,e2])
        schedule_results = scheduler.build_schedule()
        self.assertEqual(len(schedule_results), 2)
        self.assertEqual(schedule_results[0].epic_key, 'csesc-1050')
        self.assertEqual(schedule_results[0].start_date, '2023-10-25')
        self.assertEqual(schedule_results[0].end_date, '2023-10-26')
        self.assertEqual(schedule_results[0].epic_schedule_status, EpicScheduleStatus.OK)
        self.assertEqual(schedule_results[1].epic_key, 'csesc-1051')
        self.assertEqual(schedule_results[1].start_date, '2023-10-27')
        self.assertEqual(schedule_results[1].end_date, '2023-10-30')
        self.assertEqual(schedule_results[1].epic_schedule_status, EpicScheduleStatus.OK)

    def test_schedule_second_epic_wont_fit(self):
        # pylint: disable=line-too-long
        """Tests scheduling a second epic that won't fit."""
        team1_document = """
        team:
          name: Team1
          persons:
          - name: Freddy UIDev
            start_date: '2023-01-01'
            end_date: '2030-12-31'
            front_end: True
            back_end: False
            qe: False
            devops: False
            documentation: False
            reserve_capacity: 0.0
            location: US
            out_of_office_dates: []
        """
        time_period = TimePeriod(
            name='test_period',
            start_date=date(2023,10,25),
            end_date=date(2023,11,3)
        )
        e1 = Epic(
            key='csesc-1050',
            estimated_size=2,
            epic_type=EpicType.FRONTEND
        )
        e2 = Epic(
            key='csesc-1051',
            estimated_size=10,
            epic_type=EpicType.FRONTEND
        )
        team1 = Team('Team1', 'team1.yaml').load_from_yaml_as_string(team1_document)
        team1_capacity = TeamCapacity(
            team1,
            time_period,
            HolidayScheduleForTesting())
        team1_capacity.calculate()
        scheduler = TeamScheduler(team1_capacity,[e1,e2])
        schedule_results = scheduler.build_schedule()
        self.assertEqual(len(schedule_results), 2)
        self.assertEqual(schedule_results[0].epic_key, 'csesc-1050')
        self.assertEqual(schedule_results[0].start_date, '2023-10-25')
        self.assertEqual(schedule_results[0].end_date, '2023-10-26')
        self.assertEqual(schedule_results[0].epic_remaining, 0)
        self.assertEqual(schedule_results[0].epic_schedule_status, EpicScheduleStatus.OK)
        self.assertEqual(schedule_results[1].epic_key, 'csesc-1051')
        self.assertEqual(schedule_results[1].start_date, '2023-10-27')
        self.assertEqual(schedule_results[1].end_date, 'WILL NOT COMPLETE IN TIME')
        self.assertEqual(schedule_results[1].epic_remaining, 4)
        self.assertEqual(schedule_results[1].epic_schedule_status, EpicScheduleStatus.NO_CAPACITY_TO_COMPLETE)

    def test_schedule_third_epic_wont_start(self):
        # pylint: disable=line-too-long
        """Tests scheduling a third epic that won't start."""
        team1_document = """
        team:
          name: Team1
          persons:
          - name: Freddy UIDev
            start_date: '2023-01-01'
            end_date: '2030-12-31'
            front_end: True
            back_end: False
            qe: False
            devops: False
            documentation: False
            reserve_capacity: 0.0
            location: US
            out_of_office_dates: []
        """
        time_period = TimePeriod(
            name='test_period',
            start_date=date(2023,10,25),
            end_date=date(2023,10,30)
        )
        e1 = Epic(
            key='csesc-1050',
            estimated_size=2,
            epic_type=EpicType.FRONTEND
        )
        e2 = Epic(
            key='csesc-1051',
            estimated_size=2,
            epic_type=EpicType.FRONTEND
        )
        e3 = Epic(
            key='csesc-1052',
            estimated_size=2,
            epic_type=EpicType.FRONTEND
        )
        team1 = Team('Team1', 'team1.yaml').load_from_yaml_as_string(team1_document)
        team1_capacity = TeamCapacity(
            team1,
            time_period,
            HolidayScheduleForTesting())
        team1_capacity.calculate()
        scheduler = TeamScheduler(team1_capacity,[e1,e2,e3])
        schedule_results = scheduler.build_schedule()
        self.assertEqual(len(schedule_results), 3)
        self.assertEqual(schedule_results[0].epic_key, 'csesc-1050')
        self.assertEqual(schedule_results[0].start_date, '2023-10-25')
        self.assertEqual(schedule_results[0].end_date, '2023-10-26')
        self.assertEqual(schedule_results[0].epic_remaining, 0)
        self.assertEqual(schedule_results[0].epic_schedule_status, EpicScheduleStatus.OK)
        self.assertEqual(schedule_results[1].epic_key, 'csesc-1051')
        self.assertEqual(schedule_results[1].start_date, '2023-10-27')
        self.assertEqual(schedule_results[1].end_date, '2023-10-30')
        self.assertEqual(schedule_results[1].epic_remaining, 0)
        self.assertEqual(schedule_results[1].epic_schedule_status, EpicScheduleStatus.OK)
        self.assertEqual(schedule_results[2].epic_key, 'csesc-1052')
        self.assertEqual(schedule_results[2].start_date, 'WILL NOT START')
        self.assertEqual(schedule_results[2].end_date, 'WILL NOT COMPLETE IN TIME')
        self.assertEqual(schedule_results[2].epic_remaining, 2)
        self.assertEqual(schedule_results[2].epic_schedule_status, EpicScheduleStatus.NO_CAPACITY_TO_START)


"""Epic."""
from enum import Enum
from typing import NamedTuple

class EpicType(Enum):
    """Class representing the various types of supported epics."""
    FRONTEND = 1
    BACKEND = 2
    QE = 3
    DEVOPS = 4
    DOCUMENTATION = 5

class Epic(NamedTuple):
    """Class representing an epic."""
    key: str
    estimated_size: int
    epic_type: EpicType

    def to_yaml(self):
        """Converts the epic to a yaml object."""
        return {
            'key': self.key,
            'estimated_size': self.estimated_size,
            'epic_type' : self.epic_type.name
        }



"""Feature."""
import yaml
from .epic import Epic, EpicType

class Feature:
    """Class representing a feature."""

    def __init__(self, key:str) -> None:
        self.key = key
        self.epic_list: list[Epic] = []

    def add_epic(self, epic: Epic) -> None:
        """Adds an epic to the feature."""
        self.epic_list.append(epic)

    def to_yaml(self):
        """Converts the feature to a yaml."""
        return {
            'key': self.key,
            'epics': [epic.to_yaml() for epic in self.epic_list]
        }

class Features:
    """Class representing a list of features."""

    def __init__(self, file_path) -> None:
        self.feature_list: list[Feature] = []
        self.file_path = file_path

    def add_feature(self, feature: Feature) -> None:
        """Adds a feature to the list."""
        self.feature_list.append(feature)

    def load_from_yaml_as_string(self, yaml_string:str):
        """Loads features from a yaml string."""
        data = yaml.safe_load(yaml_string)
        return self.load_yaml(data)

    def load_from_yaml_file(self):
        """Loads features from a yaml file."""
        with open(self.file_path, 'r', encoding="utf-8") as file:
            data = yaml.safe_load(file)
            return self.load_yaml(data)

    def load_yaml(self,data):
        """Loads features from a yaml."""
        for feature_data in data['features']:
            feature = Feature(feature_data['key'])
            for epic_data in feature_data['epics']:
                epic = Epic(
                    epic_data['key'], 
                    epic_data['estimated_size'],
                    EpicType[epic_data['epic_type']])
                feature.add_epic(epic)
            self.feature_list.append(feature)
        return self

    def get_epics(self) -> list[Epic]:
        """Gets a list of all epics in the features."""
        l = []
        for feature in self.feature_list:
            l = l + feature.epic_list
        return l

    def to_yaml(self):
        """Writes the features to yaml file."""
        features = []
        for feature in self.feature_list:
            features.append(feature.to_yaml())
        data = {'features' :features}

        with open(self.file_path, 'w', encoding="utf-8") as file:
            yaml.dump(data, file, sort_keys=False)



"""Person class."""""
from datetime import date
from typing import NamedTuple

class Person(NamedTuple):
    """Class representing a person."""
    name: str
    start_date: date
    end_date: date
    front_end: bool
    back_end: bool
    qe: bool
    devops: bool
    documentation: bool
    reserve_capacity: float
    location: str
    out_of_office_dates: list[date]

    def is_out_of_office(self,some_date) -> bool:
        """Returns true if the person is out of office on the given date."""
        return some_date in self.out_of_office_dates
    
    def is_not_active(self, some_date) -> bool:
        """Returns true if the person is not active on the given date."""
        return self.start_date > some_date or self.end_date < some_date

    def is_not_available(self, some_date) -> bool:
        """Returns true if the person is not available on the given date."""
        return self.is_out_of_office(some_date) or self.is_not_active(some_date)

    def to_yaml(self):
        """Converts the person to yaml."""
        return {
            'name': self.name,
            'start_date': self.start_date.isoformat(),
            'end_date': self.end_date.isoformat(),
            'front_end': self.front_end,
            'back_end': self.back_end,
            'qe': self.qe,
            'devops': self.devops,
            'documentation': self.documentation,
            'reserve_capacity': self.reserve_capacity,
            'location': self.location,
            'out_of_office_dates': [d.isoformat() for d in self.out_of_office_dates],
        }

    def __str__(self) -> str:
        return f'Person: {self.name} {self.start_date} {self.end_date} {self.country}'



"""Functions for generating capacity sheets for teams and organizations."""
from .team import Team, Organization
from .teamcapacity import TeamCapacity
from .timeperiod import TimePeriod
from .holiday import HolidaySchedulePort


def generate_capacity_sheet_for_team(
        team: Team, time_period: TimePeriod,
        holiday_schedule: HolidaySchedulePort) -> TeamCapacity:
    """
    Generates a capacity sheet for a given team and time period.

    Args:
        team (Team): The team for which to generate the capacity sheet.
        time_period (TimePeriod): The time period for which to generate the capacity sheet.
        holiday_schedule (HolidaySchedulePort): The holiday schedule to use .

    Returns:
        TeamCapacity: The generated capacity sheet.
    """
    team_capacity = TeamCapacity(
        team,
        time_period,
        holiday_schedule=holiday_schedule,
    )
    team_capacity.calculate()
    df = team_capacity.get_df()
    csv_name = team.name + "_" + time_period.name + ".csv"
    df.to_csv(csv_name)
    return team_capacity

def generate_capacity_sheet_for_org(
        org_name: str,
        teams: list,
        time_period: TimePeriod,
        holiday_schedule: HolidaySchedulePort) -> TeamCapacity:
    """ Generates a capacity sheet for a given organization and time period."""
    org = Organization(org_name)
    org_team = org.generate_team(teams)
    return generate_capacity_sheet_for_team(
        org_team, time_period, holiday_schedule=holiday_schedule
    )



from .src.epic import Epic, EpicType
from .src.feature import Feature, Features
from .src.holiday import HolidaySchedulePort
from .src.person import Person
from .src.scheduler import ScheduleResult, EpicScheduleStatus, TeamScheduler
from .src.team import Team, Organization
from .src.teamcapacity import TeamCapacity
from .src.timeperiod import TimePeriod
from .src.capacitytocsv import (
    generate_capacity_sheet_for_team,
    generate_capacity_sheet_for_org
)


"""Team."""
from datetime import date
import yaml
from .person import Person

class Team:
    """Class representing a team."""
    def __init__(self, name:str, file_path) -> None:
        self.name = name
        self.file_path = file_path
        self.person_list = []

    def add_person(self, person: Person) -> None:
        """Adds a person to the team."""
        self.person_list.append(person)

    def add_list(self, list_of_people:list[Person]) -> None:
        """Adds a list of people to the team."""
        self.person_list += list_of_people

    def load_from_yaml_as_string(self, yaml_string:str):
        """Loads a team from a yaml string."""
        data = yaml.safe_load(yaml_string)
        return self.load_yaml(data)

    def load_from_yaml_file(self):
        """Loads a team from a yaml file."""
        with open(self.file_path, 'r', encoding='utf-8') as file:
            data = yaml.safe_load(file)
            return self.load_yaml(data)

    def load_yaml(self,data):
        # pylint: disable=line-too-long
        """Loads a team from yaml."""
        team_data = data['team']
        self.name = team_data['name']
        self.person_list = []
        for person_data in team_data['persons']:
            person = Person(
                name=person_data['name'],
                start_date=date.fromisoformat(person_data['start_date']),
                end_date=date.fromisoformat(person_data['end_date']),
                front_end=person_data['front_end'],
                back_end=person_data['back_end'],
                qe=person_data['qe'],
                devops=person_data['devops'],
                documentation=person_data['documentation'],
                reserve_capacity=float(person_data['reserve_capacity']),
                location=person_data['location'],
                out_of_office_dates=[date.fromisoformat(d) for d in person_data['out_of_office_dates']],
            )
            self.person_list.append(person)
        return self

    def to_yaml(self):
        """Writes the team to a yaml file."""
        persons = []
        for person in self.person_list:
            persons.append(person.to_yaml())
        team_data = {
            'name': self.name,
            'persons': persons
        }
        data = {'team' :team_data}

        with open(self.file_path, 'w', encoding='utf-8') as file:
            yaml.safe_dump(data, file, sort_keys=False)

    def __str__(self) -> str:
        return f'Team: {self.name} [ {self.person_list}]'

class Organization:
    """Class representing an organization."""

    def __init__(self, org_name:str) -> None:
        self.org_name = org_name

    def generate_team(self,teams: list[Team]) -> Team:
        """Generates a single team from a list of teams that represent the organization."""
        org_team  = Team(self.org_name, "NONE")
        for team in teams:
            org_team.add_list(team.person_list)
        return org_team



"""Scheduler."""
import copy
from typing import NamedTuple
from enum import Enum
from .teamcapacity import TeamCapacity
from .epic import EpicType, Epic


class EpicStateDuringScheduling(Enum):
    """Class representing the various states of an epic during scheduling."""
    NOT_STARTED = 1
    IN_PROGRESS = 2
    DONE = 3


class EpicScheduleStatus(Enum):
    """Class representing the outcome of a scheduled epic."""
    NO_CAPACITY_TO_START = 1
    NO_CAPACITY_TO_COMPLETE = 2
    OK = 3


class ScheduleResult(NamedTuple):
    """Class representing the result of scheduling an epic."""
    epic_schedule_status: EpicScheduleStatus
    epic_key: str
    epic_type: EpicType
    epic_estimated_size: int
    start_date: str
    end_date: str
    epic_remaining: int

    def __str__(self):
        # pylint: disable-next=line-too-long
        return f'{self.epic_key} {self.epic_estimated_size} {self.epic_type} {self.epic_schedule_status}: starts {self.start_date} and ends {self.end_date} with {self.epic_remaining} remaining'


class TeamScheduler:
    """Class for scheduling epics givwn the capacity for a team."""
    def __init__(self, team_capacity: TeamCapacity, assigned_epics: set[Epic]):
        # self.capacity_data = copy.deepcopy(team_capacity.total_capacity_data)
        self.detailed_capacity_data = copy.deepcopy(
            team_capacity.detailed_capacity_data)
        self.assigned_epics = assigned_epics

    def build_schedule(self):
        """Builds a schedule for the assigned epics."""
        schedule_results: list[ScheduleResult] = []
        for epic in self.assigned_epics:
            result = self.schedule_epic(epic)
            schedule_results.append(result)
        return schedule_results

    def who_has_capacity(self, day, category_of_work):
        """Returns a list of persons who have capacity for the category of work on the given day."""
        l = []
        for person_name in self.detailed_capacity_data[day].keys():
            remaining_capacity = sum(
                self.detailed_capacity_data[day][person_name][category_of_work])
            if remaining_capacity > 0:
                l.append(person_name)
        return l

    def apply_capacity(self, day, category_of_work, person_name, epic_remaining):
        """Applies capacity for the category of work on the given day for the given person."""
        person_has_available = sum(
            self.detailed_capacity_data[day][person_name][category_of_work])
        if person_has_available > 0:
            subtract_capacity = min(epic_remaining, person_has_available)
            epic_remaining -= subtract_capacity
            for c in EpicType:
                self.detailed_capacity_data[day][person_name][c.name].append(
                    -subtract_capacity)
        return epic_remaining

    def schedule_epic(self, epic) -> ScheduleResult:
        """Schedules an epic."""
        epic_size = epic.estimated_size
        epic_state = EpicStateDuringScheduling.NOT_STARTED
        epic_remaining = epic_size
        epic_start_date = "WILL NOT START"
        epic_end_date = "WILL NOT COMPLETE IN TIME"
        for day in self.detailed_capacity_data.keys():
            if epic_state == EpicStateDuringScheduling.DONE:
                break
            list_of_persons = self.who_has_capacity(day, epic.epic_type.name)
            for person in list_of_persons:
                if epic_state == EpicStateDuringScheduling.NOT_STARTED:
                    epic_start_date = day
                    epic_state = EpicStateDuringScheduling.IN_PROGRESS
                if epic_remaining > 0:
                    epic_remaining = self.apply_capacity(
                        day, epic.epic_type.name, person, epic_remaining)
            if epic_remaining == 0:
                epic_end_date = day
                epic_state = EpicStateDuringScheduling.DONE
        epic_schedule_status = EpicScheduleStatus.OK
        if epic_state == EpicStateDuringScheduling.NOT_STARTED:
            epic_schedule_status = EpicScheduleStatus.NO_CAPACITY_TO_START
        elif epic_state == EpicStateDuringScheduling.IN_PROGRESS:
            epic_schedule_status = EpicScheduleStatus.NO_CAPACITY_TO_COMPLETE
        return ScheduleResult(
            epic_schedule_status=epic_schedule_status,
            epic_key=epic.key,
            epic_type=epic.epic_type,
            epic_estimated_size=epic.estimated_size,
            start_date=epic_start_date,
            end_date=epic_end_date,
            epic_remaining=epic_remaining
        )



""" Setup script for agileplanner package."""
# pylint: disable=deprecated-module
from distutils.command.clean import clean
import os
from shutil import rmtree
from setuptools import setup, find_packages

class CleanCommand(clean):
    """Custom implementation of ``clean`` setuptools command."""
    CLEAN_FILES = 'dist ./app/agileplanner.egg-info'.split(' ')

    def run(self):
        """After calling the super class implementation, this function removes
        the dist directory and egg-info file if they exists."""
        super().run()
        for item_to_clean in self.CLEAN_FILES:
            print(f'Cleaning {item_to_clean}')
            if os.path.exists(item_to_clean):
                rmtree(item_to_clean)

with open("docs/README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="agileplanner",
    version="0.0.2",
    description="Agile planning tools for capacity calculation and basic scheduling of epics.",
    package_dir={"": "app"},
    packages=find_packages(where="app"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jasondchambers/agileplanner",
    author="Jason Chambers",
    author_email="jason.d.chambers@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Topic :: Office/Business :: Scheduling"
    ],
    install_requires=["pandas >= 2.1.1", "pyYAML >= 6.0.1"],
    extras_require={
        "dev": ["twine>=4.0.2"]
    },
    python_requires=">=3.10",
    cmdclass={
        'clean': CleanCommand,
    },
)



"""Unit tests for teamcapacity.py"""
import unittest
from datetime import date
from ..src.team import Team
from ..src.teamcapacity import TeamCapacity
from ..src.holiday import HolidaySchedulePort
from ..src.timeperiod import TimePeriod

class HolidayScheduleForTesting(HolidaySchedulePort):
    """HolidaySchedulePort implementation for testing."""
    # overriding abstract method
    def falls_on_holiday(self,some_date: date,location: str) -> bool:
        return False

class TestTeamCapacity(unittest.TestCase):
    """Test TeamCapacity class"""

    def test_impossible_time_period(self):
        """Test that an impossible time period raises an exception"""
        team1_document = """
        team:
          name: Team1
          persons:
          - name: Freddy UIDev
            start_date: '2023-01-01'
            end_date: '2023-01-31'
            front_end: True
            back_end: True
            qe: False
            devops: False
            documentation: False
            reserve_capacity: 0.0
            location: US
            out_of_office_dates: []
        """
        time_period = TimePeriod(
            name='test_period',
            start_date=date(2023,10,25),
            end_date=date(1990,10,25)
        )
        team1 = Team('Team1', 'team1.yaml').load_from_yaml_as_string(team1_document)
        self.assertRaises(ValueError, TeamCapacity,
            team1,
            time_period,
            HolidayScheduleForTesting())

    def test_period_greater_than_three_years(self):
        """Test that a very long time period raises an exception"""
        team1_document = """
        team:
          name: Team1
          persons:
          - name: Freddy UIDev
            start_date: '2023-01-01'
            end_date: '2023-01-31'
            front_end: True
            back_end: True
            qe: False
            devops: False
            documentation: False
            reserve_capacity: 0.0
            location: US
            out_of_office_dates: []
        """
        time_period = TimePeriod(
            name='test_period',
            start_date=date(2023,10,25),
            end_date=date(2026,10,25)
        )
        team1 = Team('Team1', 'team1.yaml').load_from_yaml_as_string(team1_document)
        self.assertRaises(ValueError, TeamCapacity,
            team1,
            time_period,
            HolidayScheduleForTesting())

    #def test_time_period_of_one_day(self): TODO
        #"""Tests a time period of one day."""
        #team1_document = """
        #team:
          #name: Team1
          #persons:
          #- name: Freddy UIDev
            #start_date: '2023-01-01'
            #end_date: '2030-12-31'
            #front_end: True
            #back_end: True
            #qe: False
            #devops: False
            #documentation: False
            #reserve_capacity: 0.0
            #location: US
            #out_of_office_dates: []
        #"""
        #time_period = TimePeriod(
            #name='test_period',
            #start_date=date(2023,10,25),
            #end_date=date(2023,10,25)
        #)
        #team1 = Team('Team1', 'team1.yaml').load_from_yaml_as_string(team1_document)
        #team1_capacity = TeamCapacity(
            #team1,
            #time_period,
            #HolidayScheduleForTesting())
        #team1_capacity.calculate()
        #df = team1_capacity.get_df()
        #self.assertEqual(df.shape[0],1) 



"""TimePeriod."""
from datetime import date
from typing import NamedTuple

class TimePeriod(NamedTuple):
    """Class representing a time period."""
    name: str
    start_date: date
    end_date: date

    def is_valid(self) -> bool:
        """Returns true if the time period is valid."""
        if self.start_date > self.end_date:
            return False
        if (self.end_date - self.start_date).days > (365 *3):
            return False
        return True

    def __str__(self) -> str:
        return f'Person: {self.name} {self.start_date} {self.end_date}'



"""HolidaySchedulePort."""
from datetime import date
from abc import ABC, abstractmethod

class HolidaySchedulePort(ABC):
    # pylint: disable=too-few-public-methods
    """Interface for a holiday schedule. A holiday schedule is a list of dates that 
    are holidays for a given location."""

    @abstractmethod
    def falls_on_holiday(self,some_date: date,location: str) -> bool:
        """Returns true if the date falls on a holiday for the given location."""
